import { Node, TransformBit, Vec3, Quat, NodeEventType, v3, mat4, Mat4, Mat3, quat, _decorator, Component } from 'cc';
const { ccclass } = _decorator;

const v3_a = v3();
const v3_b = v3();
const qt_1 = quat();
const m3_1 = new Mat3();
const m4_1 = mat4();
const m4_2 = mat4();

// override Node.setPosition()
Node.prototype.setPosition = function(val: Readonly<Vec3> | number, y?: number, z?: number) {
    if (y === undefined && z === undefined) {
        Vec3.copy(this._lpos, val as Vec3);
    } else if (z === undefined) {
        Vec3.set(this._lpos, val as number, y!, this._lpos.z);
    } else {
        Vec3.set(this._lpos, val as number, y!, z);
    }

    this.invalidateChildren(TransformBit.POSITION);
    if (this._eventMask & TRANSFORM_ON) {
        this.emit(NodeEventType.TRANSFORM_CHANGED, TransformBit.POSITION);
    }
};

// override Node.setRotation()
Node.prototype.setRotation = function(val: Readonly<Quat> | number, y?: number, z?: number, w?: number) {
    if (y === undefined || z === undefined || w === undefined) {
        Quat.copy(this._lrot, val as Readonly<Quat>);
    } else {
        Quat.set(this._lrot, val as number, y, z, w);
    }
    this._eulerDirty = true;

    this.invalidateChildren(TransformBit.ROTATION);
    if (this._eventMask & TRANSFORM_ON) {
        this.emit(NodeEventType.TRANSFORM_CHANGED, TransformBit.ROTATION);
    }
}

// override Node.setRotationFromEuler()
Node.prototype.setRotationFromEuler = function(val: Vec3 | number, y?: number, zOpt?: number) {
    const z = zOpt === undefined ? this._euler.z : zOpt;

    if (y === undefined) {
        Vec3.copy(this._euler, val as Vec3);
        Quat.fromEuler(this._lrot, (val as Vec3).x, (val as Vec3).y, (val as Vec3).z);
    } else {
        Vec3.set(this._euler, val as number, y, z);
        Quat.fromEuler(this._lrot, val as number, y, z);
    }

    this._eulerDirty = false;

    this.invalidateChildren(TransformBit.ROTATION);
    if (this._eventMask & TRANSFORM_ON) {
        this.emit(NodeEventType.TRANSFORM_CHANGED, TransformBit.ROTATION);
    }
}

// override Node.setScale
Node.prototype.setScale = function(val: Readonly<Vec3> | number, y?: number, z?: number) {
    if (y === undefined && z === undefined) {
        Vec3.copy(this._lscale, val as Vec3);
    } else if (z === undefined) {
        Vec3.set(this._lscale, val as number, y!, this._lscale.z);
    } else {
        Vec3.set(this._lscale, val as number, y!, z);
    }

    this.invalidateChildren(TransformBit.SCALE);
    if (this._eventMask & TRANSFORM_ON) {
        this.emit(NodeEventType.TRANSFORM_CHANGED, TransformBit.SCALE);
    }
}

// override Node.setWorldPosition
Node.prototype.setWorldPosition = function(val: Vec3 | number, y?: number, z?: number) {
    if (y === undefined || z === undefined) {
        Vec3.copy(this._pos, val as Vec3);
    } else {
        Vec3.set(this._pos, val as number, y, z);
    }
    const parent = this._parent;
    const local = this._lpos;
    if (parent) {
        // TODO: benchmark these approaches
        /* */
        parent.updateWorldTransform();
        Vec3.transformMat4(local, this._pos, Mat4.invert(m4_1, parent._mat));
        /* *
        parent.inverseTransformPoint(local, this._pos);
        /* */
    } else {
        Vec3.copy(local, this._pos);
    }

    this.invalidateChildren(TransformBit.POSITION);
    if (this._eventMask & TRANSFORM_ON) {
        this.emit(NodeEventType.TRANSFORM_CHANGED, TransformBit.POSITION);
    }
}

// override Node.setWorldRotation
Node.prototype.setWorldRotation = function(val: Quat | number, y?: number, z?: number, w?: number) {
    if (y === undefined || z === undefined || w === undefined) {
        Quat.copy(this._rot, val as Quat);
    } else {
        Quat.set(this._rot, val as number, y, z, w);
    }
    if (this._parent) {
        this._parent.updateWorldTransform();
        Quat.multiply(this._lrot, Quat.conjugate(this._lrot, this._parent._rot), this._rot);
    } else {
        Quat.copy(this._lrot, this._rot);
    }
    this._eulerDirty = true;

    this.invalidateChildren(TransformBit.ROTATION);
    if (this._eventMask & TRANSFORM_ON) {
        this.emit(NodeEventType.TRANSFORM_CHANGED, TransformBit.ROTATION);
    }
}

// override Node.setWorldScale
Node.prototype.setWorldScale = function(val: Vec3 | number, y?: number, z?: number) {
    const parent = this._parent;
    if (parent) {
        this.updateWorldTransform();
    }
    if (y === undefined || z === undefined) {
        Vec3.copy(this._scale, val as Vec3);
    } else {
        Vec3.set(this._scale, val as number, y, z);
    }
    if (parent) {
        v3_a.x = this._scale.x / Vec3.set(v3_b, this._mat.m00, this._mat.m01, this._mat.m02).length();
        v3_a.y = this._scale.y / Vec3.set(v3_b, this._mat.m04, this._mat.m05, this._mat.m06).length();
        v3_a.z = this._scale.z / Vec3.set(v3_b, this._mat.m08, this._mat.m09, this._mat.m10).length();
        Mat4.scale(m4_1, this._mat, v3_a);
        Mat4.multiply(m4_2, Mat4.invert(m4_2, parent._mat), m4_1);
        Mat3.fromQuat(m3_1, Quat.conjugate(qt_1, this._lrot));
        Mat3.multiplyMat4(m3_1, m3_1, m4_2);
        this._lscale.x = Vec3.set(v3_a, m3_1.m00, m3_1.m01, m3_1.m02).length();
        this._lscale.y = Vec3.set(v3_a, m3_1.m03, m3_1.m04, m3_1.m05).length();
        this._lscale.z = Vec3.set(v3_a, m3_1.m06, m3_1.m07, m3_1.m08).length();
    } else {
        Vec3.copy(this._lscale, this._scale);
    }

    this.invalidateChildren(TransformBit.SCALE);
    if (this._eventMask & TRANSFORM_ON) {
        this.emit(NodeEventType.TRANSFORM_CHANGED, TransformBit.SCALE);
    }
}

// override Node.angle
const TRANSFORM_ON = 1 << 0;
Object.defineProperty(Node.prototype, 'angle', {
    get() {
        return this.eulerAngles.z;
    },

    set(val: number) {
        Vec3.set(this._euler, 0, 0, val);
        Quat.fromAngleZ(this._lrot, val);
        this._eulerDirty = false;

        this.invalidateChildren(TransformBit.ROTATION);
        if (this._eventMask & TRANSFORM_ON) {
            this.emit(NodeEventType.TRANSFORM_CHANGED, TransformBit.ROTATION);
        }
    },

    configurable: true,
    enumerable: true,
});

@ccclass('NodeOverride')
export class NodeOverride extends Component {
    start() {

    }

    update(deltaTime: number) {
        
    }
}
